package com.android.email.mail.internet.protocol;

import android.util.Log;

import com.android.email.Email;
import com.android.email.Utility;
import com.android.email.mail.Folder;
import com.android.email.mail.ProtocolException;
import com.android.email.mail.Store;
import com.android.email.mail.internet.HttpGeneric;
import com.android.email.mail.internet.wbxml.activesync.*;
import com.android.email.mail.internet.wbxml.WBXML;
import com.android.email.mail.internet.wbxml.CodePage;
import com.android.email.mail.internet.protocol.Protocol;
import com.android.email.mail.store.ActiveSyncStore;
import com.android.email.mail.transport.TrustedSocketFactory;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Stack;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpOptions;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.entity.AbstractHttpEntity;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A class for handling all the protocol level communication and
 * data for the Microsoft Exchange ActiveSync protocol.
 *
 * @version .2
 * @author Matthew Brace
 */
public class ActiveSyncProtocol extends Protocol {
    private URI mUri;            /* Stores the URI with connection information */
    private String mUsername;    /* Stores the user supplied username for authentication */
    private String mAlias;       /* Stores just the username if the user included realm/username */
    private String mPassword;    /* Stores the user supplied password for authentication */
    private String mHost;        /* Stores the user supplied server for connection */
    private String mAuthString;  /* Stores the base64 encoded string used in headers for authentication */

    private DefaultHttpClient mHttpClient = null; /* Stores the HttpClient to be used for requests */
    private String mFolderSyncKey = "0";             /* Stores the consecutive sync key for states for folders */

    CodePage[] mCodePages;
    WBXML mWbxmlCoder;

    private static enum supportedCommands { /* Stores the supported commands */
        checkSettings, getFolderHierarchy;
    };
    
    /**
     * Public constructor.  Initializes default data and settings based on the supplied URI
     */
    public ActiveSyncProtocol(String uri) throws ProtocolException {
        /** Generate the information from the URI */
        try {
            mUri = new URI(uri);
        } catch (URISyntaxException use) {
            throw new ProtocolException("Invalid URI supplied to ActiveSyncProtocol", use);
        }

        /**
         * All requests will occur over https.  According to the spec, this is
         * the only way it can occur.
         */
        mHost = mUri.getHost();
        if (mHost.startsWith("http")) {
            String[] hostParts = mHost.split("://", 2);
            if (hostParts.length > 1) {
                mHost = hostParts[1];
            }
        }
        mHost = "https://" + mHost;

        if (mUri.getUserInfo() != null) {
            String[] userInfoParts = mUri.getUserInfo().split(":", 2);
            mUsername = userInfoParts[0];
            String userParts[] = mUsername.split("/", 2);

            if (userParts.length > 1) {
                mAlias = userParts[1];
            } else {
                mAlias = mUsername;
            }
            if (userInfoParts.length > 1) {
                mPassword = userInfoParts[1];
            }
        }

        /**
         * DefaultHttpClient managed to fail basic authentication with a result of "Bad Request (Invalid Verb)"
         * every single time in testing.  Manually generating the authentication header succeeded.
         * ActiveSync uses Basic Authentication over SSL, so generate our header based on the supplied
         * information here
         */
        mAuthString = "Basic " + Utility.base64Encode(mUsername + ":" + mPassword);
    }

    /**
     * Determines if the supplied string is a command that the protocol supports.  Command is
     * only a loose-fitting term as it is used to directly corellate to functions.
     */
    public boolean isCommandSupported(String command) {
        boolean result = false;
        try {
            supportedCommands.valueOf(command);
            result = true;
        } catch (IllegalArgumentException iae) {
            result = false;
        }
        
        return result;
    }

    /**
     * Verifies that all the supplied settings result in a successful authentication.
     * Realistically the entire object should be rebuilt if this fails, but could
     * fail due to network problems, etc.
     */
    @Override
    public boolean checkSettings() {
        boolean result = false;
        try {
            HashMap<String, String> headers = new HashMap<String, String>();
            /* No explicit headers needed */
            processRequest(mHost + "/Microsoft-Server-ActiveSync",
                           "OPTIONS",
                           null,
                           headers,
                           false,
                           "");
            result = true;
        } catch (ProtocolException pe) {
            result = false;
        }
        
        return result;
    }

    /**
     * Retrieves an HttpClient that is ready for requests.
     */
    public DefaultHttpClient getHttpClient() throws ProtocolException {
        SchemeRegistry reg;
        Scheme s;
        
        if (mHttpClient == null) {
            mHttpClient = new DefaultHttpClient();
        }

        /** Assign everything for supporting self-signed certificates */
        reg = mHttpClient.getConnectionManager().getSchemeRegistry();
        try {
            String host = mHost.replaceAll("https://", "");
            s = new Scheme("https", new TrustedSocketFactory(host, true), 443);
        } catch (NoSuchAlgorithmException nsa) {
            Log.e(Email.LOG_TAG, "NoSuchAlgorithmException in getHttpClient: " + nsa);
            throw new ProtocolException("NoSuchAlgorithmException in getHttpClient: " + nsa);
        } catch (KeyManagementException kme) {
            Log.e(Email.LOG_TAG, "KeyManagementException in getHttpClient: " + kme);
            throw new ProtocolException("KeyManagementException in getHttpClient: " + kme);
        }
        reg.register(s);

        /**
         * Return the mHttpClient.  It's the responsibility of the user of the client to
         * assign the authentication headers to the entity request object.
         */
        return mHttpClient;
    }

    /**
     * Processes the supplied request.  Throws ProtocolException if any error condition
     * is reached.  If needsParsing is false, DataSet is guaranteed to be empty.  However,
     * if needsParsing is true, DataSet is not guaranteed to be non-empty.
     */
    private DataSet processRequest(String url,
                                   String method,
                                   AbstractHttpEntity messageEntity,
                                   HashMap<String, String> headers,
                                   boolean needsParsing,
                                   String dataTag) throws ProtocolException {
        DataSet result = new DataSet(dataTag);
        DefaultHttpClient httpclient = getHttpClient();

        if (url == null) {
            throw new ProtocolException("Invalid URL supplied to processRequest");
        }
        if (method == null) {
            throw new ProtocolException("Invalid method supplied to processRequest");
        }

        try {
            int statusCode = -1;
            HttpGeneric httpmethod = new HttpGeneric(url);
            HttpResponse response;
            HttpEntity entity;

            if (messageEntity != null) {
                httpmethod.setEntity(messageEntity);
            }

            for (String headerName : headers.keySet()) {
                httpmethod.setHeader(headerName, headers.get(headerName));
            }

            /** The caller doesn't need to explicitly know anything about authentication. */
            httpmethod.setHeader("Authorization", mAuthString);

            httpmethod.setMethod(method);
            response = httpclient.execute(httpmethod);
            statusCode = response.getStatusLine().getStatusCode();

            entity = response.getEntity();

            /** 401 gets special handling to show the authentication failure */
            if (statusCode == 401) {
                throw new IOException("Authentication failed processing request");
            } else if (statusCode < 200 ||
                       statusCode > 300) {
                throw new IOException("Error during request processing: "+
                                      response.getStatusLine().toString()+ "\n\n"+
                                      getHttpRequestResponse(messageEntity, entity));
            }

            if (entity != null &&
                needsParsing) {
                try {
                    String responseContentType = response.getFirstHeader("Content-Type").getValue();
                    InputStream istream = entity.getContent();
                    if (responseContentType.startsWith("application/vnd.ms-sync")) {
                        WBXML decoder = getWbxml();
                        ByteArrayOutputStream wbxmlStream = new ByteArrayOutputStream();
                        decoder.convertWbxmlToXml(istream, wbxmlStream);
                        istream = new BufferedInputStream(new ByteArrayInputStream(wbxmlStream.toByteArray()), 200000);
                    }
                    SAXParserFactory spf = SAXParserFactory.newInstance();
                    SAXParser sp = spf.newSAXParser();
                    XMLReader xr = sp.getXMLReader();
                    ParseHandler myHandler = new ParseHandler(result.getContentTag());
                        
                    xr.setContentHandler(myHandler);
                    xr.parse(new InputSource(istream));

                    result = myHandler.getDataSet();
                } catch (SAXException se) {
                    Log.e(Email.LOG_TAG, "SAXException in processRequest() " + se + "\nTrace: " + processException(se));
                } catch (ParserConfigurationException pce) {
                    Log.e(Email.LOG_TAG, "ParserConfigurationException in processRequest() " + pce + "\nTrace: " + processException(pce));
                }
            }
        } catch (UnsupportedEncodingException uee) {
            Log.e(Email.LOG_TAG,
                  "UnsupportedEncodingException in processRequest: " + uee + "\nTrace: " + processException(uee));
            throw new ProtocolException("UnsupportedEncodingException in processRequest");
        } catch (IOException ioe) {
            Log.e(Email.LOG_TAG, "IOException in processRequest: " + ioe + "\nTrace: " + processException(ioe));
            throw new ProtocolException("IOException in processRequest");
        }

        return result;
    }

    private WBXML getWbxml() {
        if (mWbxmlCoder == null) {
            mCodePages = new CodePage[22];
            mCodePages[0] = new AirSyncCodePage();
            mCodePages[1] = new ContactsCodePage();
            mCodePages[2] = new EmailCodePage();
            mCodePages[3] = new AirNotifyCodePage();
            mCodePages[4] = new CalendarCodePage();
            mCodePages[5] = new MoveCodePage();
            mCodePages[6] = new ItemEstimateCodePage();
            mCodePages[7] = new FolderHierarchyCodePage();
            mCodePages[8] = new MeetingResponseCodePage();
            mCodePages[9] = new TasksCodePage();
            mCodePages[0xa] = new ResolveRecipientsCodePage();
            mCodePages[0xb] = new ValidateCertCodePage();
            mCodePages[0xc] = new Contacts2CodePage();
            mCodePages[0xd] = new PingCodePage();
            mCodePages[0xe] = new ProvisionCodePage();
            mCodePages[0xf] = new SearchCodePage();
            mCodePages[0x10] = new GALCodePage();
            mCodePages[0x11] = new AirSyncBaseCodePage();
            mCodePages[0x12] = new SettingsCodePage();
            mCodePages[0x13] = new DocumentLibraryCodePage();
            mCodePages[0x14] = new ItemOperationsCodePage();
            mWbxmlCoder = new WBXML(mCodePages);
        } 
        return mWbxmlCoder;
    }
    
    /**
     * Returns a string of the stacktrace for a Throwable to allow for easy inline printing of errors.
     */
    private String processException(Throwable t) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(baos);

        t.printStackTrace(ps);
        ps.close();

        return baos.toString();
    }

    /**
     * Retrieves details about an HTTP Request response.  Used to provide more detailed log information.
     */
    public static String getHttpRequestResponse(HttpEntity request, HttpEntity response) throws IllegalStateException, IOException{
        String responseText = "";
        String requestText = "";
        if (response != null) {
            BufferedReader reader = new BufferedReader(new InputStreamReader(response.getContent()), 8192);
            String tempText = "";

            while ((tempText = reader.readLine()) != null) {
                responseText += tempText;
            }
        }
        if (request != null) {
            BufferedReader reader = new BufferedReader(new InputStreamReader(request.getContent()), 8192);
            String tempText = "";

            while ((tempText = reader.readLine()) != null) {
                requestText += tempText;
            }
            requestText = requestText.replaceAll("Basic .*", "Basic (omitted)&");
        }
        return "Request: " + requestText +
            "\n\nResponse: " + responseText;

    }

    @Override
    public Folder[] getFolderHierarchy(Store store) {
        ActiveSyncStore astore = (ActiveSyncStore) store;
        String xmlData = ASXmlRequests.getFolderSyncXml(mFolderSyncKey);
        ArrayList<Folder> folderList = new ArrayList<Folder>();
        HashMap<String, String> headers = new HashMap<String, String>();
        DataSet resultSet = new DataSet("");
        ByteArrayOutputStream dataIn = new ByteArrayOutputStream();
        WBXML encoder = getWbxml();
        ByteArrayEntity messageEntity;

        encoder.convertXmlToWbxml(new ByteArrayInputStream(xmlData.getBytes()),
                                  dataIn);
        try {
            HashMap<String, String> resultData;
            HashMap<String, String>[] resultSetArray;
            dataIn.flush();

            messageEntity = new ByteArrayEntity(dataIn.toByteArray());
            messageEntity.setContentType("application/vnd.ms-sync.wbxml");

            headers.put("MS-ASProtocolVersion", "2.5");
            headers.put("Content-Type", "application/vnd.ms-sync.wbxml");
            resultSet = processRequest(mHost + "/Microsoft-Server-ActiveSync?User="+mUsername+"&DeviceId=6F24CAD599A5BF1A690246B8C68FAE8D&DeviceType=PocketPC&Cmd=FolderSync",
                                       "POST",
                                       messageEntity,
                                       headers,
                                       true,
                                       "Add");

            resultData = resultSet.getResultData();
            resultSetArray = resultSet.getResultArray();
            if (resultData.containsKey("Status") &&
                resultData.get("Status").equals("1")) {
                for (int i = 0, count = resultSetArray.length; i < count; i++) {
                    ActiveSyncStore.ActiveSyncFolder folder = astore.new ActiveSyncFolder(resultSetArray[i].get("DisplayName"),
                                                                                          resultSetArray[i].get("ServerId"));
                    folderList.add(folder);
                }
            }
        } catch (IOException ioe) {
            Log.e(Email.LOG_TAG, "IOException in getFolderHierarchy: " + ioe + "\nTrace: " + processException(ioe));
        } catch (ProtocolException pe) {
            Log.e(Email.LOG_TAG, "ProtocolException in getFolderHierachy: " + pe + "\nTrace: " + processException(pe));
        } catch (Exception e) {
            Log.e(Email.LOG_TAG, "Exception in getFolderHierarchy: " + e + "\nTrace: " + processException(e));
        }

        return folderList.toArray(new Folder[] {});
    }

    /**
     * Final class with private constructor so it cannot be instantiated.  Holds, processes and
     * creates the XML requests for ActiveSync.
     */
    public static class ASXmlRequests {
        /** This is never called */
        private ASXmlRequests() {

        }

        /**
         * Returns the XML string for the command FolderSync with the supplied syncKey.
         */
        public static String getFolderSyncXml(String syncKey) {
            String xml = "<?xml version=\"1.0\" encoding=\"utf-8\"?><FolderSync xmlns=\"FolderHierarchy\"><SyncKey>";
            xml = xml + syncKey + "</SyncKey></FolderSync>";
            return xml;
        }
    }
    
    /**
     * Dataset for all XML parses.
     * There are two sources of data that are stored.
     * 1) An array of HashMaps of tags and characters contained between open and close tags for
     * the supplied tag.
     * 2) A HashMap of tags and characters contained between those tags that don't fall within
     * the elements of the supplied tag.
     * For entries that don't require an array of HashMaps for the result, an empty string can
     * be used.
     */
    public class DataSet {
        String contentTag = new String();
        boolean insideContentTag = false;
        ArrayList<HashMap> resultSet;
        HashMap<String, String> resultData;
        HashMap<String, String> tempResultSet;
        Stack<String> tagStack;
        
        public DataSet(String tag) {
            contentTag = tag;
            resultSet = new ArrayList<HashMap>();
            resultData = new HashMap<String, String>();
            tempResultSet = new HashMap<String, String>();
            tagStack = new Stack<String>();
        }

        public String getContentTag() {
            return contentTag;
        }
        
        public void push(String tag) {
            if (tag.equals(contentTag)) {
                insideContentTag = true;
            }

            tagStack.push(tag);
        }

        public void pop() {
            if (tagStack.peek().equals(contentTag)) {
                resultSet.add(tempResultSet);
                insideContentTag = false;
                tempResultSet = new HashMap<String, String>();
            }

            tagStack.pop();
        }

        public void addCharacters(char[] ch, int start, int end) {
            if (insideContentTag) {
                tempResultSet.put(tagStack.peek(), new String(ch, start, end));
            } else {
                resultData.put(tagStack.peek(), new String(ch, start, end));
            }
        }

        public HashMap[] getResultArray() {
            HashMap<String, String>[] result = new HashMap[resultSet.size()];
            result = resultSet.toArray(result);
            return result;
        }

        public HashMap getResultData() {
            return resultData;
        }
    }
    
    /** 
     * XML Parsing Handler
     * Can handle all XML handling needs
     */
    public class ParseHandler extends DefaultHandler {
        private DataSet mDataSet;
        private String mContentTag = "";
        
        public DataSet getDataSet() {
            return this.mDataSet;
        }

        public ParseHandler(String contentTag) {
            super();
            mContentTag = contentTag;
        }
        
        @Override
        public void startDocument() throws SAXException {
            this.mDataSet = new DataSet(mContentTag);
        }

        @Override
        public void endDocument() throws SAXException {
            /* Do nothing */
        }

        @Override
        public void startElement(String namespaceURI, String localName,
                                 String qName, Attributes atts) throws SAXException {
            mDataSet.push(localName);
        }

        @Override
        public void endElement(String namespaceURI, String localName, String qName) {
            mDataSet.pop();
        }

        @Override
        public void characters(char ch[], int start, int length) {
            mDataSet.addCharacters(ch, start, length);
        }
    }
}
